/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "maya/maya.pmh"

#include "maya/MFnPlugin.h"

#define FE_MN_COMPOUND_TEST		FALSE
#define FE_MN_CATALOG_DEBUG		FALSE

// TODO may sure this is a unique name
static const char* feMayaDefaultOutputSurface = "Output Surface";

namespace fe
{

I32 mayaDiscoverOperators(MObject a_object)
{
	MFnPlugin plugin(a_object,PLUGIN_COMPANY,"1.0","Any");

	MayaContext mayaContext;

	feLog("mayaDiscoverOperators master=%p\n",
		mayaContext.master().raw());

	sp<Registry> spRegistry=mayaContext.master()->registry();
	mayaContext.loadLibraries();

	int count=0;

	std::vector<String> available;
	spRegistry->listAvailable("OperatorSurfaceI",available);
	U32 candidates=available.size();
	for(U32 index=0;index<candidates;index++)
	{
		feLog("checking %d/%d \"%s\"\n",
				index,candidates,available[index].c_str());

		sp<OperatorSurfaceI> spOperatorSurfaceI=
				spRegistry->create(available[index]);
		if(!spOperatorSurfaceI.isValid())
		{
			feLog("mayaDiscoverOperators"
					" Component is not an OperatorSurfaceI \"%s\"\n",
					available[index].c_str());
			continue;
		}

		String implementation=available[index];
		char* buffer1=strchr(const_cast<char*>(available[index].c_str()),'.');
		if(buffer1)
		{
			buffer1++;
			const char* buffer2=strchr(buffer1,'.');
			if(buffer2>buffer1)
			{
				implementation.sPrintf("%.*s",int(buffer2-buffer1),buffer1);
			}
			else
			{
				implementation=buffer1;
			}
		}

		feLog("mayaDiscoverOperators implementation \"%s\"\n",
				implementation.c_str());

		const U32 uniqueId=0x80013+count;	//* HACK

		const char* name=strdup(implementation.c_str());
		const char* english=strdup(("FE_"+implementation).c_str());

		fe::MayaNode::preRegister(uniqueId,english,available[index]);
		MStatus status=plugin.registerNode(
				english,
				MTypeId(uniqueId),
				fe::MayaNode::create,
				fe::MayaNode::initialize);
		if(!status)
		{
			status.perror("registerNode");
		}
		fe::MayaNode::postRegister();

		count++;
	}

	return count;
}

} // namespace fe

extern "C"
{

//* discovery point for Maya
FE_DL_EXPORT MStatus initializePlugin(MObject a_object)
{
	feLog("::initializePlugin\n");

	I32 count=fe::mayaDiscoverOperators(a_object);

	feLog("::initializePlugin done\n");

	return MS::kSuccess;
}

FE_DL_EXPORT MStatus uninitializePlugin(MObject a_object)
{
	feLog("::uninitializePlugin\n");

	MFnPlugin plugin(a_object);

	//* TODO proper
/*
	MStatus status=plugin.deregisterNode(fe::MayaNode::ms_id);
	if(!status)
	{
		status.perror("deregisterNode");
	}
*/

	feLog("::uninitializePlugin done\n");

	return MS::kSuccess;
}

//* TODO Is this unneeded in Maya?
#if FALSE
FE_DL_EXPORT void HoudiniDSOInit(UT_DSOInfo& dsoinfo)
{
	//* force Houdini to load with RTLD_GLOBAL (needed for dynamic_cast<>)
	dsoinfo.loadGlobal=true;
}
#endif

}

namespace fe
{

I32	MayaNode::ms_registerLock=0;
String MayaNode::ms_registerName;
MObject	MayaNode::ms_compoundBehavior;
std::map<String,MayaNode::Registration> MayaNode::ms_mayaRegistry;

MayaNode::MayaNode(void):
	m_lastTime(0.0f)
{
	feLog("MayaNode::MayaNode master=%p\n",m_mayaContext.master().raw());

	sp<Registry> spRegistry=m_mayaContext.master()->registry();
	m_mayaContext.loadLibraries();

	m_spDrawI=spRegistry->create("DrawI.MayaDraw");
	feLog("MayaNode::MayaNode m_spDrawI valid=%d\n",m_spDrawI.isValid());

	sp<Scope> spScope=m_mayaContext.scope();

	m_spSignalerI=spRegistry->create("SignalerI");
	m_spSignalerI->setName("MayaNode.SignalerI");

	m_aSurfaceOutput.setup(spScope,"SurfaceOutput","component");
	m_aDrawI.setup(spScope,"DrawI","component");

	m_spLayout = spScope->declare("MayaNode");
	m_spLayout->populate(m_aSurfaceOutput);
	m_spLayout->populate(m_aDrawI);

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(MayaNode),"MayaNode");
#endif

	feLog("MayaNode::MayaNode done\n");
}

MayaNode::~MayaNode(void)
{
	feLog("~MayaNode\n");

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif
}

void MayaNode::preRegister(U32 a_uniqueId,String a_name,
		String a_implementation)
{
	feLog("MayaNode::preRegister\n");
	FEASSERT(ms_registerLock==0);
	ms_registerLock++;
	ms_registerName=a_name;

	Registration& registration=ms_mayaRegistry[a_name];
	registration.m_autoOutput=FALSE;
	registration.m_uniqueId=a_uniqueId;
	registration.m_implementation=a_implementation;
}

void MayaNode::postRegister(void)
{
	feLog("MayaNode::postRegister\n");
	FEASSERT(ms_registerLock==1);

	ms_registerName="";
	ms_registerLock--;
}

MStatus MayaNode::initialize(void)
{
	feLog("MayaNode::initialize %s %d\n",
			ms_registerName.c_str(),ms_registerLock);

	populateOperator(ms_registerName);

	feLog("MayaNode::initialize done\n");

	return MS::kSuccess;
}

void* MayaNode::create(void)
{
	feLog("MayaNode::create\n");
	return new MayaNode();
}

void MayaNode::populateOperator(String a_name)
{
	MayaContext mayaContext;

	feLog("MayaNode:populateOperators master=%p\n",
		mayaContext.master().raw());

	sp<Registry> spRegistry=mayaContext.master()->registry();

	Registration& registration=ms_mayaRegistry[a_name];
	String implementation=registration.m_implementation;
	std::map<String,Parameter>& rParameterMap=registration.m_parameterMap;

	feLog("registration \"%s\" \"%s\"\n",
			a_name.c_str(),implementation.c_str());

	sp<OperatorSurfaceI> spOperatorSurfaceI=
			spRegistry->create(implementation);
	if(!spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode:populateOperators"
				" Component is not an OperatorSurfaceI \"%s\"\n",
				implementation.c_str());
		return;
	}

	//* tried to move built-in attributes to a compound
	MStatus status=MS::kSuccess;
#if FE_MN_COMPOUND_TEST
//	MFnAttribute fnCaching;
//	fnCaching.setObject(caching);
//	fnCaching.setHidden(true);

//	MFnAttribute fnState;
//	fnState.setObject(state);
//	fnState.setHidden(true);

	MFnCompoundAttribute fnBehavior;
	ms_compoundBehavior=fnBehavior.create("NodeBehavior","behavior",&status);
	status=fnBehavior.setNiceNameOverride("Node Behavior");
	FEASSERT(status);
//	status=fnBehavior.addChild(caching);
	FEASSERT(status);
//	status=fnBehavior.addChild(state);
	FEASSERT(status);
#endif

	std::vector<String> keys;

	sp<Catalog> spCatalog=spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:populateOperators"
				" Component is not a Catalog \"%s\"\n",
				implementation.c_str());

		//* TODO keep this from crashing later
	}
	else
	{
		spCatalog->catalogKeys(keys);
	}

	U32 number=keys.size();
	feLog("MayaNode::populateOperators keys %d\n",number);

	U32 outputs=0;
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		if(!spCatalog->catalogLookup(keys[m],instance))
		{
			continue;
		}
		if(instance.is< sp<Component> >())
		{
			char* keyname=strdup(keys[m].c_str());
			String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");
			if(io=="output")
			{
				outputs++;
			}
		}
	}

	if(!outputs)
	{
		feLog("MayaNode::populateOperators no output surface specified\n");

		//* refresh key list
		keys.clear();
		registration.m_autoOutput=TRUE;

		if(spCatalog.isValid())
		{

			spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
			spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")=
					"output";

			spCatalog->catalogKeys(keys);
			number=keys.size();
		}
	}

	for(U32 m=0;m<number;m++)
	{
		feLog("MayaNode::populateOperators key %d: \"%s\"\n",
				m,keys[m].c_str());

		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::populateOperators invalid Instance\n");
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());
		String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");

		if(instance.is< sp<Component> >())
		{
			//* see meshOpNode.cpp

			String implementation=
					spCatalog->catalogOrDefault<String>(keyname,
					"implementation",
					"SurfaceAccessibleI.SurfaceAccessibleMaya");

			if(implementation=="SurfaceAccessibleI" || implementation==
					"SurfaceAccessibleI.SurfaceAccessibleMaya")
			{
				MFnTypedAttribute typedAttr;
				rParameterMap[keyname].m_object=typedAttr.create(
						cleanname,cleanname,MFnMeshData::kMesh);
				typedAttr.setNiceNameOverride(keyname);
				if(io=="input")
				{
					U32 channel=registration.m_inputChannelArray.size();
					registration.m_inputChannelArray.resize(channel+1);
					registration.m_inputChannelArray[channel]=keyname;

					typedAttr.setStorable(true);
				}
				else
				{
					U32 channel=registration.m_outputChannelArray.size();
					registration.m_outputChannelArray.resize(channel+1);
					registration.m_outputChannelArray[channel]=keyname;

					typedAttr.setStorable(false);
					typedAttr.setWritable(false);
				}
				MStatus status=addAttribute(rParameterMap[keyname].m_object);
			}
		}
		else
		{
			MFnAttribute attribute;

			if(instance.is<String>())
			{
				char* value=strdup(instance.cast<String>().c_str());

				MFnStringData fnStringData;
				MObject defaultString=fnStringData.create(value);
				MFnTypedAttribute stringAttr;
				rParameterMap[keyname].m_object=
						stringAttr.create(cleanname,cleanname,
						MFnData::kString,defaultString);
				attribute.setObject(stringAttr.object());
			}
			else
			{
				MFnNumericAttribute numAttr;

				if(instance.is<bool>())
				{
					bool value=instance.cast<bool>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kBoolean,value);
				}
				else if(instance.is<Real>())
				{
					Real value=instance.cast<Real>();

					Real min=spCatalog->catalogOrDefault<Real>(keyname,
							"min",0.0);
					Real max=spCatalog->catalogOrDefault<Real>(keyname,
							"max",1.0);
					Real low=spCatalog->catalogOrDefault<Real>(keyname,
							"low",min);
					Real high=spCatalog->catalogOrDefault<Real>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kDouble,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);
				}
				else if(instance.is<I32>())
				{
					I32 value=instance.cast<I32>();

					I32 min=spCatalog->catalogOrDefault<I32>(keyname,
							"min",0);
					I32 max=spCatalog->catalogOrDefault<I32>(keyname,
							"max",10);
					I32 low=spCatalog->catalogOrDefault<I32>(keyname,
							"low",min);
					I32 high=spCatalog->catalogOrDefault<I32>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kInt,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);
				}
				else if(instance.is<SpatialVector>())
				{
					SpatialVector& value=instance.cast<SpatialVector>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k3Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2]);
				}
				else if(instance.is<Vector4>())
				{
					Vector4& value=instance.cast<Vector4>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k4Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2],(double)value[3]);
				}

				attribute.setObject(numAttr.object());
			}

			attribute.setNiceNameOverride(keyname);
			if(io=="input")
			{
				attribute.setKeyable(true);
				attribute.setStorable(true);
			}
			else
			{
				attribute.setKeyable(false);
				attribute.setStorable(false);
			}

#if FE_MN_COMPOUND_TEST
			MStatus status=
					fnBehavior.addChild(rParameterMap[keyname].m_object);
#else
			MStatus status=addAttribute(rParameterMap[keyname].m_object);
#endif
		}
	}

#if FE_MN_COMPOUND_TEST
	status=addAttribute(ms_compoundBehavior);
	FEASSERT(status);
#endif

	//* Currently every output is dependent on every input.
	//* We could get more specific.
	std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
	while(it!=rParameterMap.end())
	{
		const String& key=it->first;
		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		if(io=="output")
		{
			const Parameter& parameter=it->second;
			const MObject& rParamOutput=parameter.m_object;

			std::map<String,Parameter>::const_iterator it2=
					rParameterMap.begin();
			while(it2!=rParameterMap.end())
			{
				const String& key2=it2->first;
				String io2=
						spCatalog->catalogOrDefault<String>(key2,"IO","input");
				if(io2=="input")
				{
					const Parameter& parameter2=it2->second;
					const MObject& rParamInput=parameter2.m_object;

					MStatus status=attributeAffects(rParamInput,rParamOutput);
					if(!status)
					{
						feLog("MayaNode::populateOperator"
								" attributeAffects %s -> %s failed\n",
								key2.c_str(),key.c_str());
					}
				}
				it2++;
			}
		}
		it++;
	}

	feLog("MayaNode::populateOperator done\n");
}

void MayaNode::updateCatalog(Real time,BWORD a_doInput,MDataBlock& rData)
{
#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog doInput=%d\n",a_doInput);
#endif

	if(!m_spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode::updateCatalog invalid m_spOperatorSurfaceI\n");
		return;
	}

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:updateCatalog operator is not a Catalog\n");
		return;
	}

	Registration& registration=ms_mayaRegistry[m_name];

	BWORD autocopyInput=FALSE;
	if(registration.m_autoOutput)
	{
#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateCatalog auto-output\n");
#endif

		spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
		spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")="output";
		autocopyInput=TRUE;
	}

	const std::map<String,Parameter>& rParameterMap=registration.m_parameterMap;

	std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
	while(it!=rParameterMap.end())
	{
		const String& key=it->first;
		const Parameter& parameter=it->second;
		const MObject& rObject=parameter.m_object;

#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateCatalog \"%s\"\n",key.c_str());
#endif

		Instance instance;
		bool success=spCatalog->catalogLookup(key,instance);
		if(!success)
		{
			feLog("MayaNode::updateCatalog invalid Instance \"%s\"\n",
					key.c_str());
			it++;
			continue;
		}

		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		BWORD isInput=(io=="input");
		if(isInput!=a_doInput)
		{
//			feLog("MayaNode::updateCatalog skip \"%s\"\n",io.c_str());
			it++;
			continue;
		}

		if(instance.is<String>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					spCatalog->catalog<String>(key)=
							handleIn.asString().asChar();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(MString(spCatalog->catalog<String>(key).c_str()));
			}
		}
		else if(instance.is<bool>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					spCatalog->catalog<bool>(key)=handleIn.asBool();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(spCatalog->catalog<bool>(key));
			}
		}
		else if(instance.is<Real>())
		{
			Real& rValue=spCatalog->catalog<Real>(key);

			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asDouble();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(double(rValue));
			}
#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now Real %.6G\n",rValue);
#endif
		}
		else if(instance.is<I32>())
		{
			I32& rValue=spCatalog->catalog<I32>(key);

			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asInt();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rValue);
			}
#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now I32 %d\n",rValue);
#endif
		}
		else if(instance.is<SpatialVector>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					double3& rValue=handleIn.asDouble3();
					set(spCatalog->catalog<SpatialVector>(key),rValue[0],
							rValue[1],rValue[2]);
				}
			}
			else
			{
				SpatialVector& rVector=spCatalog->catalog<SpatialVector>(key);

				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rVector[0],rVector[1],rVector[2]);
			}
		}
		else if(instance.is<Vector4>())
		{
			// TODO four floats is oddly supported
		}
		else if(instance.is< sp<Component> >())
		{
			//* meshes handled elsewhere
		}

		it++;
	}

	sp<Registry> spRegistry=m_mayaContext.master()->registry();

	std::vector<String> keys;
	spCatalog->catalogKeys(keys);

	const U32 number=keys.size();
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::updateCatalog invalid Instance \"%s\"\n",
					keys[m].c_str());
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());

		if(instance.is< sp<Component> >())
		{
			sp<Component>& rspComponent=instance.cast< sp<Component> >();

			if(!rspComponent.isValid())
			{
				String implementation=spCatalog->catalogOrDefault<String>(
						keyname,"implementation",
						"SurfaceAccessibleI.SurfaceAccessibleMaya");
				if(!implementation.empty())
				{
					if(implementation=="SurfaceAccessibleI")
					{
						implementation=
								"SurfaceAccessibleI.SurfaceAccessibleMaya";
					}
					rspComponent=spRegistry->create(implementation);
				}
				sp<SurfaceAccessibleI> spSurfaceAccessibleI=rspComponent;
				if(spSurfaceAccessibleI.isValid())
				{
					String io=spCatalog->catalogOrDefault<String>(
							keyname,"IO","input");
					if(io=="input")
					{
						I32 channel=m_surfaceInputArray.size();
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateCatalog"
								" adding surface input %d \"%s\" \"%s\"\n",
								channel,cleanname,keyname);
#endif
						spCatalog->catalog<I32>(keyname,"channel")=channel;
						m_surfaceInputArray.push_back(spSurfaceAccessibleI);

						if(autocopyInput)
						{
							spCatalog->catalog<String>(
									feMayaDefaultOutputSurface,"copy")=keyname;
							autocopyInput=FALSE;
						}
					}
					else
					{
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateCatalog"
								" adding surface output \"%s\" \"%s\"\n",
								cleanname,keyname);
#endif
						m_spSurfaceOutput=spSurfaceAccessibleI;

						// HACK
						m_nameOutput=keyname;
					}
				}
			}
		}
	}

#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog done\n");
#endif
}

MStatus MayaNode::compute(const MPlug& rPlug,MDataBlock& rData)
{
	feLog("MayaNode::compute \"%s\"\n",rPlug.name().asChar());

	// TODO check maya node state (normal, passthrough, etc)

	m_name=typeName().asChar();
	feLog("  name \"%s\"\n",m_name.c_str());
	Registration& rRegistration=ms_mayaRegistry[m_name];
	const U32 uniqueId=rRegistration.m_uniqueId;
	feLog("  unique ID %p\n",uniqueId);
	m_implementation=rRegistration.m_implementation;
	feLog("  implementation \"%s\"\n",m_implementation.c_str());
	std::map<String,Parameter>& rParameterMap=rRegistration.m_parameterMap;

	double time=0.0;			// TODO

	String operatorName=m_implementation;

	if(!m_spOperatorSurfaceI.isValid())
	{
		sp<Registry> spRegistry=m_mayaContext.master()->registry();
		m_spOperatorSurfaceI=spRegistry->create(operatorName);
		feLog("MayaNode::compute m_spOperatorSurfaceI \"%s\" valid=%d\n",
				operatorName.c_str(),m_spOperatorSurfaceI.isValid());

		if(m_spOperatorSurfaceI.isValid())
		{
			updateCatalog(time,TRUE,rData);

			m_lastName=operatorName;

			if(m_spSignalerI.isValid())
			{
				m_spSignalerI->insert(m_spOperatorSurfaceI, m_spLayout);
			}
		}
	}
	else
	{
		updateCatalog(time,TRUE,rData);
	}

	if(m_spOperatorSurfaceI.isValid())
	{
		double frame=0;	// TODO

		sp<SurfaceAccessibleMaya> spSurfaceOutput=m_spSurfaceOutput;
		if(!spSurfaceOutput.isValid())
		{
			sp<Registry> spRegistry=m_mayaContext.master()->registry();
			spSurfaceOutput=spRegistry->create(
					"SurfaceAccessibleI.SurfaceAccessibleMaya");
		}
		feLog("MayaNode::compute spSurfaceOutput valid=%d\n",
				spSurfaceOutput.isValid());

		I32 channel=0;
		if(!m_nameOutput.empty())
		{
			sp<Catalog> spCatalog=m_spOperatorSurfaceI;
			String copy=spCatalog->catalogOrDefault<String>(m_nameOutput,
					"copy","");
			channel= -1;
			if(!copy.empty())
			{
				channel=spCatalog->catalogOrDefault<I32>(copy,"channel",-1);
			}
		}

		// TODO multiple outputs with individual channels?

		String outputName=rRegistration.m_outputChannelArray[0];
		MObject paramOutMesh=rParameterMap[outputName].m_object;

		MStatus status=MS::kSuccess;
		MDataHandle outputData=rData.outputValue(paramOutMesh,&status);
		FEASSERT(status);
		MObject meshOutput=outputData.asMesh();

		MDataHandle channelData;

		U32 surfaces=m_surfaceInputArray.size();
		FEASSERT(surfaces==rRegistration.m_inputChannelArray.size());
		for(int surface=0;surface<surfaces;surface++)
		{
			sp<SurfaceAccessibleMaya> spSurfaceInput=
					m_surfaceInputArray[surface];

			String inputName=rRegistration.m_inputChannelArray[surface];
			MObject paramInMesh=rParameterMap[inputName].m_object;

			MStatus status=MS::kSuccess;

			MDataHandle inputData=rData.inputValue(paramInMesh,&status);
			FEASSERT(status);
			MObject meshInput=inputData.asMesh();

			spSurfaceInput->setMeshData(inputData);

			if(surface==channel)
			{
				channelData=inputData;
			}

#if FALSE
			// HACK alternate attribute test
			sp<SurfaceAccessorI> spAccessor=spSurfaceInput->accessor(
					SurfaceAccessibleI::e_point,"milliIndex");
			FEASSERT(spAccessor.isValid());

			if(spAccessor.isValid())
			{
				const I32 pointCount=spAccessor->count();
				for(I32 index=0;index<pointCount;index++)
				{
//					feLog("milli %d/%d\n",index,spAccessor->count());
					spAccessor->set(index,0,Real(0.001*index));
				}
			}
#endif
		}

		if(channel>=0 && channel<surfaces)
		{
			feLog("MayaNode::compute output=input\n");

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			MFnMesh newMeshFn;
			newMeshFn.copy(channelData.asMesh(),newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute mesh copy failed: %s\n",
						status.errorString().asChar());
				return status;
			}
			feLog("MayaNode::compute newMesh blind used: %d %d %d\n",
					newMeshFn.isBlindDataTypeUsed(0x0),
					newMeshFn.isBlindDataTypeUsed(0x1),
					newMeshFn.isBlindDataTypeUsed(0x2));

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();
		}
		else
		{
			// erase all prior geometry

			feLog("MayaNode::compute output reset\n");

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			int numVertices=0;
			int numPolygons=0;
			MFloatPointArray vertexArray;
			MIntArray polygonCounts;
			MIntArray polygonConnects;

			status=MS::kSuccess;
			MFnMesh newMeshFn;
			MObject mesh=newMeshFn.create(numVertices,numPolygons,
					vertexArray,polygonCounts,polygonConnects,
					newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute create failed: %s\n",
						status.errorString().asChar());
			}

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();
		}
		spSurfaceOutput->setMeshData(outputData);

		sp<MayaDraw> spMayaDraw=m_spDrawI;
		if(spMayaDraw.isValid())
		{
			feLog("MayaNode::compute set draw mesh\n");
			spMayaDraw->setMeshObject(meshOutput);
		}

		Record signal = m_mayaContext.scope()->createRecord(m_spLayout);
		m_aSurfaceOutput(signal)=spSurfaceOutput;
		m_aDrawI(signal)=m_spDrawI;
		m_spSignalerI->signal(signal);

		updateCatalog(time,FALSE,rData);
	}

	m_lastTime=time;

	feLog("MayaNode::compute done\n");

	return MS::kSuccess;
}

} // namespace
